@@ -2,33 +2,48 @@ use crate::config::{Config, Delimiter};
 use crate::select::SelectColumns;
 use crate::util;
 use crate::CliResult;
-use cached::proc_macro::cached;
+use cached::proc_macro::{cached, io_cached};
+use cached::RedisCache;
+use cached::Return;
 use indicatif::{ProgressBar, ProgressDrawTarget};
 use log::{debug, error};
+use once_cell::sync::Lazy;
 use serde::Deserialize;
+use thiserror::Error;
 
 static USAGE: &str = "
-Fetch data via a URL column, and optionally store them in a new column.
-
-This command fetches data from web api for every row in the URL column, 
+This command fetches data from a web API for every row in the URL column, 
 and optionally stores them in a new column.
 
-Fetch is integrated with `jql` to directly parse out values from api JSON response.
+Fetch is integrated with `jql` to directly parse out values from API JSON response.
 
-URL column must contain full and valid URL path, which can be constructed via the `lua` command.
+URL column can either be a fully qualified URL path, or if not, can be used with
+the --url-template option to create one.
 
-To set proxy, please set env var HTTP_PROXY and HTTPS_PROXY (eg export HTTPS_PROXY=socks5://127.0.0.1:1086)
+To use a proxy, please set env vars HTTP_PROXY and HTTPS_PROXY
+(e.g. export HTTPS_PROXY=socks5://127.0.0.1:1086).
+
+Set the --redis flag to use Redis. By default, it will connect to a local Redis instance 
+at redis://127.0.0.1:6379, with a cache expiry TTL of 2,419,200 seconds (28 days), and
+cache hits NOT refreshing the TTL of cached values.
+
+Set the env vars QSV_REDIS_CONNECTION_STRING, QSV_REDIS_TTL_SECONDS and 
+QSV_REDIS_TTL_REFRESH to change default settings.
 
 Usage:
     qsv fetch [options] [--http-header <k:v>...] [<column>] [<input>]
 
 Fetch options:
+    --url-template <template>  URL template to use. The character '^'
+                               will be replaced by <column> value, but
+                               sanitized for shell safety.
     -c, --new-column <name>    Put the fetched values in a new column instead.
     --jql <selector>           Apply jql selector to API returned JSON value.
-    --rate-limit <qps>         Rate Limit in Queries Per Second. [default: 5]
+    --rate-limit <qps>         Rate Limit in Queries Per Second. [default: 10]
     --http-header <key:value>  Pass custom header(s) to the server.
     --store-error              On error, store error code/message instead of blank value.
     --cookies                  Allow cookies.
+    --redis                    Use Redis to cache responses.
 
 Common options:
     -h, --help                 Display this message
@@ -44,12 +59,14 @@ Common options:
 
 #[derive(Deserialize, Debug)]
 struct Args {
+    flag_url_template: Option<String>,
     flag_new_column: Option<String>,
     flag_jql: Option<String>,
     flag_rate_limit: Option<u32>,
     flag_http_header: Vec<String>,
     flag_store_error: bool,
     flag_cookies: bool,
+    flag_redis: bool,
     flag_output: Option<String>,
     flag_no_headers: bool,
     flag_delimiter: Option<Delimiter>,
@@ -58,6 +75,9 @@ struct Args {
     arg_input: Option<String>,
 }
 
+static DEFAULT_REDIS_CONN_STR: &str = "redis://127.0.0.1:6379";
+static DEFAULT_REDIS_TTL_SECONDS: u64 = 60 * 60 * 24 * 28; // 28 days in seconds
+
 static DEFAULT_USER_AGENT: &str = concat!(
     env!("CARGO_PKG_NAME"),
     "/",
@@ -65,6 +85,33 @@ static DEFAULT_USER_AGENT: &str = concat!(
     " (https://github.com/jqnatividad/qsv)",
 );
 
+struct RedisConfig {
+    conn_str: String,
+    ttl_secs: u64,
+    ttl_refresh: bool,
+}
+impl RedisConfig {
+    fn load() -> Self {
+        Self {
+            conn_str: std::env::var("QSV_REDIS_CONNECTION_STRING")
+                .unwrap_or_else(|_| DEFAULT_REDIS_CONN_STR.to_string()),
+            ttl_secs: std::env::var("QSV_REDIS_TTL_SECS")
+                .unwrap_or_else(|_| DEFAULT_REDIS_TTL_SECONDS.to_string())
+                .parse()
+                .unwrap(),
+            ttl_refresh: std::env::var("QSV_REDIS_TTL_REFRESH").is_ok(),
+        }
+    }
+}
+
+static REDISCONFIG: Lazy<RedisConfig> = Lazy::new(RedisConfig::load);
+
+#[derive(Error, Debug, PartialEq, Clone)]
+enum FetchRedisError {
+    #[error("error with redis cache `{0}`")]
+    RedisError(String),
+}
+
 pub fn run(argv: &[&str]) -> CliResult<()> {
     let args: Args = util::get_args(USAGE, argv)?;
 
@@ -87,7 +134,7 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
 
     use std::num::NonZeroU32;
     // default rate limit is actually set via docopt, so below init is just to satisfy compiler
-    let mut rate_limit: NonZeroU32 = NonZeroU32::new(5).unwrap();
+    let mut rate_limit: NonZeroU32 = NonZeroU32::new(10).unwrap();
     if let Some(qps) = args.flag_rate_limit {
         assert!(
             // on my laptop, no more sleep traces with qps > 24, so use round number of 20 as single-thread qps limit
@@ -143,8 +190,9 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
 
     // prep progress bar
     let progress = ProgressBar::new(0);
+    let mut record_count = 0;
     if !args.flag_quiet {
-        let record_count = util::count_rows(&rconfig);
+        record_count = util::count_rows(&rconfig);
         util::prep_progress(&progress, record_count);
     } else {
         progress.set_draw_target(ProgressDrawTarget::hidden());
@@ -152,6 +200,9 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
 
     #[allow(unused_assignments)]
     let mut record = csv::ByteRecord::new();
+    #[allow(unused_assignments)]
+    let mut url = String::new();
+    let mut redis_cache_hits: u64 = 0;
     while rdr.read_byte_record(&mut record)? {
         if !args.flag_quiet {
             progress.inc(1);
@@ -161,16 +212,37 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
         let mut final_value = String::default();
 
         if let Ok(s) = std::str::from_utf8(&record[column_index]) {
-            let url = s.trim().to_string();
+            match args.flag_url_template {
+                Some(ref url_template) => {
+                    url = url_template.replace('^', s.trim());
+                }
+                _ => url = s.trim().to_string(),
+            }
+
             debug!("Fetching URL: {url:?}");
 
-            final_value = get_cached_response(
-                &url,
-                &client,
-                &limiter,
-                &args.flag_jql,
-                args.flag_store_error,
-            );
+            if args.flag_redis {
+                let intermediate_value = get_redis_response(
+                    &url,
+                    &client,
+                    &limiter,
+                    &args.flag_jql,
+                    args.flag_store_error,
+                )
+                .unwrap();
+                final_value = intermediate_value.to_string();
+                if intermediate_value.was_cached {
+                    redis_cache_hits += 1;
+                }
+            } else {
+                final_value = get_cached_response(
+                    &url,
+                    &client,
+                    &limiter,
+                    &args.flag_jql,
+                    args.flag_store_error,
+                );
+            }
         } else {
             final_value = "Invalid URL".to_string();
         }
@@ -186,7 +258,11 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
     }
 
     if !args.flag_quiet {
-        util::update_cache_info!(progress, GET_CACHED_RESPONSE);
+        if args.flag_redis {
+            util::update_cache_info!(progress, redis_cache_hits, record_count);
+        } else {
+            util::update_cache_info!(progress, GET_CACHED_RESPONSE);
+        }
         util::finish_progress(&progress);
     }
 
@@ -200,7 +276,7 @@ use std::{thread, time};
 
 #[cached(
     key = "String",
-    convert = r#"{ format!("{}", url) }"#,
+    convert = r#"{ format!("{}{:?}", url, flag_jql) }"#,
     sync_writes = false
 )]
 fn get_cached_response(
@@ -209,6 +285,47 @@ fn get_cached_response(
     limiter: &governor::RateLimiter<NotKeyed, InMemoryState, DefaultClock, NoOpMiddleware>,
     flag_jql: &Option<String>,
     flag_store_error: bool,
+) -> String {
+    get_response(url, client, limiter, flag_jql, flag_store_error)
+}
+
+#[io_cached(
+    type = "cached::RedisCache<String, String>",
+    key = "String",
+    convert = r#"{ format!("{}{:?}", url, flag_jql) }"#,
+    create = r##" {
+        RedisCache::new("f", REDISCONFIG.ttl_secs)
+            .set_namespace("q")
+            .set_refresh(REDISCONFIG.ttl_refresh)
+            .set_connection_string(&REDISCONFIG.conn_str)
+            .build()
+            .expect("error building redis cache")
+    } "##,
+    map_error = r##"|e| FetchRedisError::RedisError(format!("{:?}", e))"##,
+    with_cached_flag = true
+)]
+fn get_redis_response(
+    url: &str,
+    client: &reqwest::blocking::Client,
+    limiter: &governor::RateLimiter<NotKeyed, InMemoryState, DefaultClock, NoOpMiddleware>,
+    flag_jql: &Option<String>,
+    flag_store_error: bool,
+) -> Result<cached::Return<String>, FetchRedisError> {
+    Ok(Return::new(get_response(
+        url,
+        client,
+        limiter,
+        flag_jql,
+        flag_store_error,
+    )))
+}
+
+fn get_response(
+    url: &str,
+    client: &reqwest::blocking::Client,
+    limiter: &governor::RateLimiter<NotKeyed, InMemoryState, DefaultClock, NoOpMiddleware>,
+    flag_jql: &Option<String>,
+    flag_store_error: bool,
 ) -> String {
     // wait until RateLimiter gives Okay
     while limiter.check().is_err() {
