@@ -5,6 +5,7 @@ use crate::CliResult;
 use anyhow::{anyhow, Result};
 use csv::ByteRecord;
 use indicatif::{ProgressBar, ProgressDrawTarget};
+use jsonschema::paths::PathChunk;
 use jsonschema::{output::BasicOutput, JSONSchema};
 #[allow(unused_imports)]
 use log::{debug, info};
@@ -12,6 +13,7 @@ use rayon::prelude::*;
 use serde::Deserialize;
 use serde_json::{value::Number, Map, Value};
 use std::{env, fs::File, io::BufReader, io::BufWriter, io::Read, io::Write, str};
+use thousands::Separable;
 
 macro_rules! fail {
     ($mesg:expr) => {
@@ -29,12 +31,11 @@ Example output files from `mydata.csv`. If piped from stdin, then filename is `s
 
 * mydata.csv.valid
 * mydata.csv.invalid
-* mydata.csv.validation-errors.jsonl
+* mydata.csv.validation-errors.tsv
 
 JSON Schema can be a local file or a URL.
 
-When run without JSON Schema, only a simple CSV check (RFC 4180) is performed, with the caveat that
- on non-Windows machines, each record is delimited by a LF (\\n) instead of CRLF (\\r\\n).
+When run without JSON Schema, only a simple CSV check (RFC 4180) is performed.
 
 
 Usage:
@@ -81,12 +82,34 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
     // if no json schema supplied, only let csv reader validate csv file
     if args.arg_json_schema.is_none() {
         // just read csv file and let csv reader report problems
-        let mut record = csv::ByteRecord::new();
-        while rdr.read_byte_record(&mut record)? {
-            // this loop is for csv::reader to do basic csv validation on read
+
+        let mut header_msg = String::new();
+        if !args.flag_no_headers {
+            let header_results = rdr.headers();
+            match header_results {
+                Ok(header) => {
+                    let header_len = header.len();
+                    header_msg = format!("{} columns and ", header_len.separate_with_commas());
+                }
+                Err(e) => return fail!(format!("Cannot read header ({e}).")),
+            }
         }
 
-        let msg = "Can't validate without schema, but csv looks good.".to_string();
+        let mut record_count: u64 = 0;
+        for result in rdr.records() {
+            if let Err(e) = result {
+                return fail!(format!(
+                    r#"There's a problem with your CSV's format ({e}).\nTry "qsv fixlengths" or "qsv fmt" to fix it."#
+                ));
+            }
+            record_count += 1;
+        }
+
+        let msg = format!(
+            "{header_msg}{} records detected. Can't validate data without schema, but CSV looks good!",
+            record_count.separate_with_commas()
+        );
+
         info!("{msg}");
         println!("{msg}");
 
@@ -146,7 +169,7 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
     // reuse batch buffer
     let mut batch = Vec::with_capacity(BATCH_SIZE);
     let mut valid_flags: Vec<bool> = Vec::with_capacity(record_count as usize);
-    let mut validation_error_messages: Vec<String> = Vec::new();
+    let mut validation_error_messages: Vec<String> = Vec::with_capacity(50);
 
     // main loop to read CSV and construct batches for parallel processing.
     // each batch is processed via Rayon parallel iterator.
@@ -180,16 +203,10 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
         let batch_size = batch.len();
 
         // do actual validation via Rayon parallel iterator
+        // validation_results vector should have same row count and in same order as input CSV
         let validation_results: Vec<Option<String>> = batch
             .par_iter()
-            .map(|record| {
-                match do_json_validation(&headers, record, &schema_json, &schema_compiled) {
-                    Ok(o) => o,
-                    Err(e) => {
-                        panic!("Unrecoverable error: {e:?}");
-                    }
-                }
-            })
+            .map(|record| do_json_validation(&headers, record, &schema_json, &schema_compiled))
             .collect();
 
         batch.clear();
@@ -220,9 +237,21 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
         }
     } // end infinite loop
 
+    if !args.flag_quiet {
+        progress.set_message(format!(
+            " validated {} records.",
+            progress.length().separate_with_commas()
+        ));
+        util::finish_progress(&progress);
+    }
+
     // only write out invalid/valid/errors output files if there are actually invalid records.
     // if 100% invalid, then valid file is not needed. but this is rare so live with creating empty file.
     if invalid_count > 0 {
+        let msg = "Writing invalid/valid/error files...";
+        info!("{msg}");
+        println!("{msg}");
+
         let input_path = args
             .arg_input
             .clone()
@@ -243,16 +272,6 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
         )?;
     }
 
-    use thousands::Separable;
-
-    if !args.flag_quiet {
-        progress.set_message(format!(
-            " validated {} records.",
-            progress.length().separate_with_commas()
-        ));
-        util::finish_progress(&progress);
-    }
-
     // done with validation; print output
     if args.flag_fail_fast {
         let msg = format!(
@@ -328,10 +347,12 @@ fn write_error_report(input_path: &str, validation_error_messages: Vec<String>)
         .unwrap_or_else(|_| DEFAULT_WTR_BUFFER_CAPACITY.to_string());
     let wtr_buffer_size: usize = wtr_capacitys.parse().unwrap_or(DEFAULT_WTR_BUFFER_CAPACITY);
 
-    let output_file = File::create(input_path.to_owned() + ".validation-errors.jsonl")?;
+    let output_file = File::create(input_path.to_owned() + ".validation-errors.tsv")?;
 
     let mut output_writer = BufWriter::with_capacity(wtr_buffer_size, output_file);
 
+    output_writer.write_all("row_number\tfield\terror\n".as_bytes())?;
+
     // write out error report
     for error_msg in validation_error_messages {
         output_writer.write_all(error_msg.as_bytes())?;
@@ -345,49 +366,41 @@ fn write_error_report(input_path: &str, validation_error_messages: Vec<String>)
     Ok(())
 }
 
-/// if given record is valid, CliResult would hold None, otherwise, Some(Value)
+/// if given record is valid, return None, otherwise, error file entry string
 fn do_json_validation(
     headers: &ByteRecord,
     record: &ByteRecord,
     schema_json: &Value,
     schema_compiled: &JSONSchema,
-) -> CliResult<Option<String>> {
-    // row number was added as last column
-    let row_number_string = str::from_utf8(record.get(headers.len()).unwrap()).unwrap();
-    let row_number: usize = row_number_string.parse::<usize>().unwrap();
+) -> Option<String> {
+    // row number was added as last column. We use unsafe from_utf8_unchecked to
+    // skip UTF8 validation since we know its safe as we added it earlier
+    let row_number_string = unsafe { str::from_utf8_unchecked(record.get(headers.len()).unwrap()) };
 
     let instance: Value = match to_json_instance(headers, record, schema_json) {
         Ok(obj) => obj,
         Err(e) => {
-            return fail!(format!(
-                "Unable to convert CSV to json. row: {row_number}, error: {e}"
-            ));
+            return Some(format!("{row_number_string}\t<RECORD>\t{e}"));
         }
     };
 
     // debug!("instance[{row_number}]: {instance:?}");
 
     match validate_json_instance(&instance, schema_compiled) {
-        Ok(mut validation_result) => {
-            let is_valid = validation_result.get("valid").unwrap().as_bool().unwrap();
+        Some(validation_errors) => {
+            use itertools::Itertools;
+            // squash multiple errors into one long String with linebreaks
+            let combined_errors: String = validation_errors
+                .iter()
+                .map(|tuple| {
+                    // validation error file format: row_number, field, error
+                    format!("{row_number_string}\t{}\t{}", tuple.0, tuple.1)
+                })
+                .join("\n");
 
-            if is_valid {
-                Ok(None)
-            } else {
-                // debug!("row[{row_number}] is invalid");
-
-                // insert row index and return validation error
-                validation_result.as_object_mut().unwrap().insert(
-                    "row_number".to_string(),
-                    Value::Number(Number::from(row_number)),
-                );
-
-                Ok(Some(validation_result.to_string()))
-            }
-        }
-        Err(e) => {
-            return fail!(format!("Unable to validate. row: {row_number}, error: {e}"));
+            Some(combined_errors)
         }
+        None => None,
     }
 }
 
@@ -399,7 +412,8 @@ fn to_json_instance(headers: &ByteRecord, record: &ByteRecord, schema: &Value) -
         .expect("JSON Schema missing 'properties' object");
 
     // map holds individual CSV fields converted as serde_json::Value
-    let mut json_object_map: Map<String, Value> = Map::new();
+    // we use with_capacity to minimize allocs
+    let mut json_object_map: Map<String, Value> = Map::with_capacity(50);
 
     // iterate over each CSV field and convert to JSON type
     let headers_iter = headers.iter().enumerate();
@@ -608,14 +622,39 @@ mod tests_for_csv_to_json_conversion {
 }
 
 /// Validate JSON instance against compiled JSON schema
-fn validate_json_instance(instance: &Value, schema_compiled: &JSONSchema) -> Result<Value> {
-    let output: BasicOutput = schema_compiled.apply(instance).basic();
+/// If invalid, returns Some(Vec<(String,String)>) holding the error messages
+fn validate_json_instance(
+    instance: &Value,
+    schema_compiled: &JSONSchema,
+) -> Option<Vec<(String, String)>> {
+    let validation_output = schema_compiled.apply(instance);
 
-    match serde_json::to_value(output) {
-        Ok(json) => Ok(json),
-        Err(e) => Err(anyhow!(
-            "Cannot convert schema validation output to json: {e}"
-        )),
+    // If validation output is Invalid, then grab field names and errors
+    if !validation_output.flag() {
+        // get validation errors as String
+        let validation_errors: Vec<(String, String)> = match validation_output.basic() {
+            BasicOutput::Invalid(errors) => errors
+                .iter()
+                .map(|e| {
+                    if let Some(PathChunk::Property(box_str)) = e.instance_location().last() {
+                        (box_str.to_string(), e.error_description().to_string())
+                    } else {
+                        (
+                            e.instance_location().to_string(),
+                            e.error_description().to_string(),
+                        )
+                    }
+                })
+                .collect(),
+            BasicOutput::Valid(_annotations) => {
+                // shouln't happen
+                unreachable!("Unexpected error.");
+            }
+        };
+
+        Some(validation_errors)
+    } else {
+        None
     }
 }
 
@@ -668,9 +707,9 @@ mod tests_for_schema_validation {
 
         let instance = to_json_instance(&headers, &record, &schema_json()).unwrap();
 
-        let result = validate_json_instance(&instance, &compiled_schema()).unwrap();
+        let result = validate_json_instance(&instance, &compiled_schema());
 
-        assert_eq!(true, result["valid"].as_bool().unwrap());
+        assert!(result.is_none());
     }
 
     #[test]
@@ -685,9 +724,17 @@ mod tests_for_schema_validation {
 
         let instance = to_json_instance(&headers, &record, &schema_json()).unwrap();
 
-        let result = validate_json_instance(&instance, &compiled_schema()).unwrap();
+        let result = validate_json_instance(&instance, &compiled_schema());
 
-        assert_eq!(false, result["valid"].as_bool().unwrap());
+        assert!(result.is_some());
+
+        assert_eq!(
+            vec![(
+                "name".to_string(),
+                "\"X\" is shorter than 2 characters".to_string()
+            )],
+            result.unwrap()
+        );
     }
 }
 
