@@ -5,6 +5,7 @@ use std::io::{self, Read};
 use std::ops::Deref;
 use std::path::PathBuf;
 
+use encoding_rs_io::DecodeReaderBytes;
 use serde::de::{Deserialize, Deserializer, Error};
 
 use crate::index::Indexed;
@@ -298,7 +299,10 @@ impl Config {
         Ok(match self.path {
             None => Box::new(io::stdin()),
             Some(ref p) => match fs::File::open(p) {
-                Ok(x) => Box::new(x),
+                Ok(x) => {
+                    let transcoded = DecodeReaderBytes::new(x);
+                    Box::new(transcoded)
+                }
                 Err(err) => {
                     let msg = format!("failed to open {}: {}", p.display(), err);
                     return Err(io::Error::new(io::ErrorKind::NotFound, msg));
@@ -307,6 +311,7 @@ impl Config {
         })
     }
 
+    #[allow(clippy::wrong_self_convention)]
     pub fn from_reader<R: Read>(&self, rdr: R) -> csv::Reader<R> {
         let rdr_capacitys = env::var("QSV_RDR_BUFFER_CAPACITY")
             .unwrap_or_else(|_| DEFAULT_RDR_BUFFER_CAPACITY.to_string());
@@ -335,6 +340,7 @@ impl Config {
         })
     }
 
+    #[allow(clippy::wrong_self_convention)]
     pub fn from_writer<W: io::Write>(&self, wtr: W) -> csv::Writer<W> {
         let wtr_capacitys = env::var("QSV_WTR_BUFFER_CAPACITY")
             .unwrap_or_else(|_| DEFAULT_WTR_BUFFER_CAPACITY.to_string());
